#!/bin/bash

set -e         # stop on errors
QUIT="q"       # the syntax to quit a pass
FILE="${1%.c}" # The first argument is the file
ARG="$2"       # The snd argument indicates the first section to start.
REPLAY="$3"    # A non-zero on the third arg indicates to replay (without recompiling)
CNT=0          # number of the current pass

function simple(){
    CNT=$((CNT+1))
    local NAME="${CNT} -- $1: $2"
    if [ -n "${ARG}" ] && [ "${ARG}" -le "${CNT}" ]; then
        read -s -p "${NAME}" -n 1 line
        echo
        if [ "${line}" != "${QUIT}" ]; then
            (echo "/* ${NAME} */"; cat $2) | less
        fi
    else
        echo "${NAME}"
    fi
}

function double(){
    CNT=$((CNT+1))
    local NAME="${CNT} -- $1: $2"
    if [ -n "${ARG}" ] && [ "${ARG}" -le "${CNT}" ]; then
        read -s -p "${NAME}" -n 1 line
        echo
        if [ "${line}" != "${QUIT}" ]; then
            (echo "/* ${NAME} */"; diff -y "$3" "$2") | less
        fi
    else
        echo "${NAME}"
    fi
}

if [ ! -f "${FILE}.c" ]; then
    echo "ERROR: \"${FILE}.c\" inexistant file"
    echo
fi

if [ ! -f "${FILE}.c" ] || [ -z "${ARG}" ]; then
    cat <<EOF
Usage: $0 <file>.c <num> [replay]
       where <file>.c is the file to compile with CompCert
       and <num> is one of the pass number below
       and optional [replay] argument to avoid recompilation

This will interactively show successive CompCert passes from pass <num>.
Type '${QUIT}' to skip or quit each pass screen (or Ctrl-C to abort all),
or arrows to scroll up/down each screen.

Example: the command below will present all passes from the source.
        $0 ${FILE}.c 1
       
EOF
    ARG=""
elif [ -z "${REPLAY}" ]; then
    echo "Compiling ${FILE}.c with CompCert"
    ccomp -O -S -dall "${FILE}.c"
    rm *.i *.parsed.c *.compcert.c *.cm
    sed -e "s/() {/:/g" -e "s/}//g" < "${FILE}.mach" > "${FILE}.machx"
fi
    echo
echo "-- Main passes of CompCert:"
simple "source C file" "${FILE}.c"
double "side-effects & calls apart from expressions" "${FILE}.light.c" "${FILE}.c"
double "initial CFG" "${FILE}.rtl.0" "${FILE}.light.c"
#double "tailcall detection" "${FILE}.rtl.1" "${FILE}.rtl.0"
double "function inlining" "${FILE}.rtl.2" "${FILE}.rtl.0"
double "CFG renumbering" "${FILE}.rtl.3" "${FILE}.rtl.2"
double "constant propagation" "${FILE}.rtl.4" "${FILE}.rtl.3"
double "CFG renumbering" "${FILE}.rtl.5" "${FILE}.rtl.4"
double "common subexp elimination" "${FILE}.rtl.6" "${FILE}.rtl.5"
double "redundancy elimination" "${FILE}.rtl.8" "${FILE}.rtl.6"
simple "oracle allocating registers" "${FILE}.alloctrace"
double "register allocation" "${FILE}.ltl" "${FILE}.rtl.8"
double "stacking registers" "${FILE}.mach" "${FILE}.ltl"
double "target assembly" "${FILE}.s" "${FILE}.machx"
